/// \file Matrix.h
/// \date 2007/11/13
/// \author Peter Dodds
/// \brief 
/// \note Copyright (C) 2007 - All Rights Reserved
//------------------------------------------------------------------------------

#ifdef _MSC_VER
#pragma once
#endif

#ifndef _MCG_Matrix_h_
#define _MCG_Matrix_h_

#include "Def.h"
#include <iostream>

namespace MCG
{
  struct Quaternion;
  struct Vector3;

#pragma pack(push,1) // do i need this? (probably not)

  /// \brief structure to hold a 4x4 matrix, in openGL form
  struct Matrix
  {
    //---------------------------------------
    /// \brief default constructor
    //Matrix();
    //---------------------------------------
    /// \brief constructor taking individual matrix elements
    /// \note in the openGL form (ie. translation along the bottom)
    Matrix(Real x0 = 1, Real y0 = 0, Real z0 = 0, Real w0 = 0,
           Real x1 = 0, Real y1 = 1, Real z1 = 0, Real w1 = 0,
           Real x2 = 0, Real y2 = 0, Real z2 = 1, Real w2 = 0,
           Real tx = 0, Real ty = 0, Real tz = 0, Real tw = 1);
    //---------------------------------------
    /// \brief constructor taking another matrix
    Matrix(const Matrix& m);
    //---------------------------------------
    /// \brief constructor taking a vector to create a translation matrix
    Matrix(const Vector3& translation);
    //---------------------------------------
    /// \brief constructor taking a quaternion to create a rotation matrix
    Matrix(const Quaternion& rotation);
    //---------------------------------------
    /// \brief destructor
    ~Matrix();

    //---------------------------------------
    /// \brief load the identity matrix
    /// \return "void" 
    void Identity();

    //---------------------------------------
    /// \brief transpose (flip along parallel) this matrix
    /// \return "const Matrix&" the transposed matrix
    const Matrix& Transpose();
    //---------------------------------------
    /// \brief get the transpose (flip along parallel) of this matrix
    /// \return "MCG::Matrix" the transposed matrix
    Matrix GetTranspose() const;

    //---------------------------------------
    /// \brief calculate the determinant of this matrix
    /// \return "MCG::Real" the determinant
    Real Determinant() const;
    //---------------------------------------
    /// \brief invert this matrix
    /// \return "const Matrix&" the inverted matrix
    const Matrix& Inverse();
    //---------------------------------------
    /// \brief get inverse of this matrix
    /// \return "MCG::Matrix" the inverted matrix
    Matrix GetInverse() const;

    Quaternion ToQuat ();

    Matrix operator * (const Matrix& m) const;
    Vector3 operator * (const Vector3& v) const;
    const Matrix& operator *= (const Matrix& m);
    friend const Vector3& operator *= (Vector3 &v,const Matrix &m);

    Matrix operator * (const Real& d) const;
    Matrix operator / (const Real& d) const;
    const Matrix& operator *= (const Real& d);
    const Matrix& operator /= (const Real& d);

		const Matrix& operator = (const Matrix& m);
    const Matrix& operator = (Quaternion& q);

		bool operator == (const Matrix& m) const;
		bool operator != (const Matrix& m) const;

    friend std::ostream& operator << (std::ostream &s, const Matrix& m);
    
    union
    { // using the openGL format (ie. translations on the bottom)
      Real m_data[16];
      struct
      {
        Real m00, m01, m02, m03; // |x0 y0 z0 w0|
        Real m10, m11, m12, m13; // |x1 y1 z1 w1|
        Real m20, m21, m22, m23; // |x2 y2 z2 w2|
        Real m30, m31, m32, m33; // |tx ty tz tw|
      };
    };
  };
#pragma pack(pop)
}

#endif
