/// \file AiAgent.h
/// \date 2008/01/24
/// \author Peter Dodds
/// \brief 
/// \note Copyright (C) 2008 - All Rights Reserved
//------------------------------------------------------------------------------

#ifdef _MSC_VER
#pragma once
#endif

#ifndef _BUG_AiAgent_h_
#define _BUG_AiAgent_h_

#include "AiDef.h"
#include <vector>

/// \brief Operation types for Ai_AddMove order queue
enum
{
  k_InsertEnd, ///< push order onto the end of the queue
  k_InsertStart, ///< insert order at the start of the queue
  k_Replace ///< remove all element in the queue then push order onto queue
};

namespace BUG
{
  class AiPath;
  class AiMove;

	/// \brief A AI agent data type. This is used to represent a agent in BUG.
  /// To create an agent, inherit from this class
	class AiAgent
	{
	public:
		//---------------------------------------
		/// \brief Default Constructor, adds pointer to this class to AiManager.
		AiAgent();
		//---------------------------------------
		/// \brief Destructor, removes pointer from AiManager
		virtual ~AiAgent();

    // -----------------------------------
    /// \brief Get the agents unique ID
    /// \returns "unsigned"
    unsigned GetID () { return m_id; }

    // -----------------------------------
    /// \brief Update the agent. This processes the move order queue.
    /// Sends each AiMove order to Ai_ProcessMove
    /// \param "const unsigned dt" that time passed since last update
    void Ai_Update (const unsigned dt);
    
    // -----------------------------------
    /// \brief This adds a AiMove order of type into the position specified by op
    /// \param "const BUGu8 & type" the type of AiMove
    /// \param "const BUGu8 & op" the type of insertion operation
    /// \returns "AiMove*" the newly added Move (use this to add specific data)
    virtual AiMove* Ai_AddMove (const BUGu8& type, const BUGu8& op = k_InsertEnd);

    // -----------------------------------
    /// \brief Convert a AiPath into a series of AiMove_Goto, and add them to the move queue.
    /// \param "const AiPath & path" The path to be followed
    /// \param "const BUGu8 & op" the type of insertion operation
    void Ai_UsePathAsMoves (const AiPath& path, const BUGu8& op = k_Replace);

  protected:
    // define BUG controls for the agent
    // -----------------------------------
    /// \brief Abstract translation method to be implemented in derived class
    /// \param "const Real & _x" x translation
    /// \param "const Real & _y" y translation
    virtual void Translate (const Real& _x, const Real& _y) = 0;
    // -----------------------------------
    /// \brief Abstract rotation method to be implemented in derived class
    /// \param "const Real & _r" angle to rotate about.
    virtual void Rotate (const Real& _r) = 0;
    // -----------------------------------
    /// \brief Abstract get translation method to be implemented in derived class
    /// \param "Real & _x" returned x translation
    /// \param "Real & _y" returned y translation
    /// \returns "void"
    virtual void GetTranslation (Real& _x, Real& _y) = 0;
    // -----------------------------------
    /// \brief Abstract rotation method to be implemented in derived class
    /// \param "Real & _r" returned rotation
    virtual void GetRotation (Real& _r) = 0;

    // functions to be replaced if there are non-default orders/states
    // -----------------------------------
    /// \brief Process classes derived from AiMove.
    /// this function should be implemented in derived classes to process custom move orders
    /// \param "AiMove * _o" move to be processed
    /// \param "const unsigned dt" that time passed since last update
    /// \returns "bool" if able the order was successfully processed
    virtual bool Ai_ProcessMove (AiMove* _o, const unsigned dt);
    // -----------------------------------
    /// \brief Create a new move. This allocates the memory for derived AiMove classes.
    /// To add custom AiMoves, implement this function.
    /// The type should be processed, and the the matching data type should be used to allocate memory.
    /// Then the pointer to this memory should be passed back.
    /// \param "BUGu8 type" the type of derived AiMove class
    /// \returns "AiMove*" the newly allocated AiMove class
    virtual AiMove* Ai_NewMove (BUGu8 type);

    unsigned m_id; ///< unique ID
    std::vector<AiMove*> m_move_q; ///< move order queue
    unsigned m_move_age; ///< time when order at the front came to the front

  private:
    // -----------------------------------
    /// \brief generate a unique ID for a new AiAgent.
    /// \returns "unsigned" the new ID
    static unsigned GetUniqueID () {return _nextID++;}
    static unsigned _nextID; ///< the next ID GetUniqueID will return
	};
}

#endif