/// \file AiBlock.cpp
/// \date 2008/03/03
/// \author Peter Dodds
/// \brief 
/// \note Copyright (C) 2008 - All Rights Reserved
//------------------------------------------------------------------------------

#include "AiBlock.h"

namespace BUG
{
  //------------------------------------------------ AiBlock::AiBlock
  AiBlock::AiBlock( BUGu8 _type /*= BUG_Block_NULL*/ ) : type(_type)
  {

  }

  //------------------------------------------------ AiBlock::~AiBlock
  AiBlock::~AiBlock()
  {

  }

  //------------------------------------------------ AiBlock::IsPointInside_Block
  bool AiBlock::IsPointInside_Block( const Real& x, const Real& y )
  {
    Real nx = x - m_pos.x, ny = y - m_pos.y;
    return nx > m_min.x && ny > m_min.y && nx < m_max.x && ny < m_max.y;
  }

  //------------------------------------------------ AiBlock::IsPointInside
  bool AiBlock::IsPointInside( const Real& x, const Real& y )
  {
    return IsPointInside_Block(x,y);
  }

  //------------------------------------------------ AiBlock_Circle::AiBlock_Circle
  AiBlock_Circle::AiBlock_Circle() : AiBlock (BUG_Block_Circle)
  {

  }

  //------------------------------------------------ AiBlock_Circle::~AiBlock_Circle
  AiBlock_Circle::~AiBlock_Circle()
  {

  }

  //------------------------------------------------ AiBlock_Circle::UpdateBoundingBox
  void AiBlock_Circle::UpdateBoundingBox()
  {
    m_min.x = m_pos.x - radius;
    m_min.y = m_pos.y - radius;

    m_max.x = m_pos.x + radius;
    m_max.y = m_pos.y + radius;
  }

  //------------------------------------------------ AiBlock_Circle::IsPointInside
  bool AiBlock_Circle::IsPointInside( const Real& x, const Real& y )
  {
    if (IsPointInside_Block(x,y))
      return IsPointInside_Circle(x,y);
    return false;
  }

  //------------------------------------------------ AiBlock_Circle::IsPointInside_Circle
  bool AiBlock_Circle::IsPointInside_Circle( const Real& x, const Real& y )
  {
    Real nx = x - m_pos.x, ny = y - m_pos.y;
    return (nx*nx + ny*ny) < (radius*radius);
  }

  //------------------------------------------------ AiBlock_Polygon::AiBlock_Polygon
  AiBlock_Polygon::AiBlock_Polygon() : AiBlock (BUG_Block_Polygon)
  {

  }

  //------------------------------------------------ AiBlock_Polygon::~AiBlock_Polygon
  AiBlock_Polygon::~AiBlock_Polygon()
  {

  }

  //------------------------------------------------ AiBlock_Polygon::PushVertex
  bool AiBlock_Polygon::PushVertex( const Real& x, const Real& y )
  {
    // add the vertex.
    Vector v;
    v.x = x;
    v.y = y;

    vertices.push_back(v);

    if (vertices.size() > 3 && !IsValid())
    {
      //vertices.pop_back();
      //return false;
    }

    UpdateBoundingBox();
    return true;
  }

  //------------------------------------------------ AiBlock_Polygon::UpdateBoundingBox
  void AiBlock_Polygon::UpdateBoundingBox()
  {
    std::vector<Vector>::iterator it = vertices.begin();
    std::vector<Vector>::iterator itend = vertices.end();

    m_min = m_max = *it;
    ++it;

    for (;it!=itend;++it)
    {
      if ((*it).x < m_min.x)
        m_min.x = (*it).x;
      else if ((*it).x > m_max.x)
        m_max.x = (*it).x;


      if ((*it).y < m_min.y)
        m_min.y = (*it).y;
      else if ((*it).y > m_max.y)
        m_max.y = (*it).y;
    }

    CalculateNormals();
  }

	//------------------------------------------------ AiBlock_Polygon::IsValid
  bool AiBlock_Polygon::IsValid()
  {
    if (vertices.size() < 3)
      return false;

    // check the first point is convex
    Real dist = DistanceToLine(vertices.back(), vertices[1], vertices[0]);
    bool outside_polygon = dist > 0;

    std::vector<Vector>::iterator it = vertices.begin() + 1;
    std::vector<Vector>::iterator itend = vertices.end() - 1;
    for (;it!=itend;++it)
    {
      dist = DistanceToLine( *(it-1),*(it+1),*it );

      // the side-ed-ness has changed (right -> left, or left -> right)
      if ((dist > 0) != outside_polygon)
        return false;
    }

    dist = DistanceToLine(*(it-1), vertices[0], *it);
    // the side-ed-ness has changed (right -> left, or left -> right)
    if ((dist > 0) != outside_polygon)
      return false;

    positive_order = outside_polygon;
    return true;
  }

  //------------------------------------------------ AiBlock_Polygon::IsPointInside
  bool AiBlock_Polygon::IsPointInside( const Real& x, const Real& y )
  {
    if (IsPointInside_Block(x,y))
      return IsPointInside_Polygon(x,y);
    return false;
  }

  //------------------------------------------------ AiBlock_Polygon::IsPointInside_Polygon
  bool AiBlock_Polygon::IsPointInside_Polygon( const Real& x, const Real& y )
  {
    Vector pnt; pnt.x = x - m_pos.x; pnt.y = y - m_pos.y;

    std::vector<Vector>::iterator it = vertices.begin();
    std::vector<Vector>::iterator itend = vertices.end();
    std::vector<Vector>::iterator n_it = normals.begin();
    std::vector<Vector>::iterator n_itend = normals.end();
    for (;it!=itend;++it, ++n_it)
    {
      Real dist = GetDistanceFromLine(*it, *n_it, pnt);//DistanceToLine(*it, *n_it, pnt);

      //Real _2nd_dist = DistanceToLine(*it, *(it+1), pnt);
      bool less = (dist < 0);
      bool valid = (less && positive_order  || !less && !positive_order);
      //if (!valid)
      if ((dist < 0) != positive_order)
        return false;
    }

    return true;
  }

  //------------------------------------------------ AiBlock_Polygon::GetDistanceFromLine
  BUG::Real AiBlock_Polygon::GetDistanceFromLine( const Vector& line_point, const Vector& line_normal, const Vector& test_point )
  {
    /*
    | d
    |---.test_point
    |  ^
    n^ =      | / R
    line_normal |/
    <--.line_point
    |
    |

    d = n^.R
    */

    return (test_point.x - line_point.x) * line_normal.x + (test_point.y - line_point.y) * line_normal.y;
  }

  //------------------------------------------------ AiBlock_Polygon::CalculateNormals
  void AiBlock_Polygon::CalculateNormals()
  {
    if (vertices.size() < 2)
      return;

    normals.clear();

    std::vector<Vector>::iterator it = vertices.begin();
    std::vector<Vector>::iterator itend = vertices.end() - 1;
    for (;it!=itend;++it)
    {
      Vector _n ( -((*(it+1)).y - (*it).y), (*(it+1)).x - (*it).x );
      Real mag = BUG_SQRT(_n.x * _n.x + _n.y * _n.y);
      if (mag == 0)
        mag = 0.00001f;
      mag = 1/mag;

      _n.x *= mag;
      _n.y *= mag;

      normals.push_back(_n);
    }

    Vector _n ( -(vertices[0].y - (*it).y), vertices[0].x - (*it).x );
    Real mag = BUG_SQRT(_n.x * _n.x + _n.y * _n.y);
    if (mag == 0)
      mag = 0.00001f;
    mag = 1/mag;

    _n.x *= mag;
    _n.y *= mag;

    normals.push_back(_n);
  }
}