/// \file AiCollision.h
/// \date 2008/03/01
/// \author Peter Dodds
/// \brief 
/// \note Copyright (C) 2008 - All Rights Reserved
//------------------------------------------------------------------------------

#ifdef _MSC_VER
#pragma once
#endif

#ifndef _BUG_AiBlock_h_
#define _BUG_AiBlock_h_

#include "AiDef.h"
#include <vector>

/// \brief possible types of AiBlock
enum
{
  BUG_Block_NULL, ///< just the bounding box

  BUG_Block_Circle, ///< circle object
  BUG_Block_Polygon, ///< convex polygon object

  BUG_Block_User_Enum ///< user expandable enum value
};

namespace BUG
{
	/// \brief The basic collidable object. 
  /// This contains the type information and bounding box test.
	class AiBlock
	{
	public:
		//---------------------------------------
		/// \brief Default Constructor
		/// \param "BUGu8 _type" used by derived classes to identify themselves
		AiBlock(BUGu8 _type = BUG_Block_NULL);
		//---------------------------------------
		/// \brief Destructor
		virtual ~AiBlock();

    // -----------------------------------
    /// \brief Get the type of a derived AiBlock
    /// \returns "BUGu8" the type
    BUGu8 GetType () const { return type; }

    // -----------------------------------
    /// \brief Interface to test if a point is within any AiBlock
    /// \param "const Real & x" point x position
    /// \param "const Real & y" point y position
    /// \returns "bool" true if inside the bounding box
    virtual bool IsPointInside (const Real& x, const Real& y);
    
    Vector m_pos; ///< absolute position

    Vector m_min; ///< relative minimum x and y of bounding box
    Vector m_max; ///< relative maximum x and y of bounding box 

  private:
    BUGu8 type;

  protected:
    // -----------------------------------
    /// \brief Test if point is within the bounding box
    /// \param "const Real & x" point x position
    /// \param "const Real & y" point y position
    /// \returns "bool" true if inside the bounding box
    bool IsPointInside_Block (const Real& x, const Real& y);
	};

  /// \brief Circle object derived from AiBlock
  class AiBlock_Circle : public AiBlock
  {
  public:
    //---------------------------------------
    /// \brief Default Constructor, setting type to BUG_Block_Circle
    AiBlock_Circle();
    //---------------------------------------
    /// \brief Destructor
    virtual ~AiBlock_Circle();

    // -----------------------------------
    /// \brief Update the bounding box minimums and maximums to fit circle radius.
    void UpdateBoundingBox ();

    // -----------------------------------
    /// \brief Interface to test if a point is within any AiBlock
    /// \param "const Real & x" point x position
    /// \param "const Real & y" point y position
    /// \returns "bool" true if inside the bounding box
    virtual bool IsPointInside (const Real& x, const Real& y);

    Real radius; ///< circle radius

  protected:
    // -----------------------------------
    /// \brief Test if point is within the circle
    /// \param "const Real & x" point x position 
    /// \param "const Real & y" point y position
    /// \returns "bool" true if inside the bounding box
    bool IsPointInside_Circle (const Real& x, const Real& y);
  };

  /// \brief Convex Polygon object derived from AiBlock
  class AiBlock_Polygon : public AiBlock
  {
  public:
    //---------------------------------------
    /// \brief Default Constructor, setting type to BUG_Block_Polygon
    AiBlock_Polygon();
    //---------------------------------------
    /// \brief Destructor
    virtual ~AiBlock_Polygon();

    // -----------------------------------
    /// \brief Add Vertex to polygon. 
    /// Done in order, if point is not valid (the shape becomes concave) then it is ignored.
    /// \param "const Real & x" point x position 
    /// \param "const Real & y" point y position
    /// \returns "bool" true if successfully added
    bool PushVertex (const Real& x, const Real& y);
    // -----------------------------------
    /// \brief Update bounding box to match the polygon
    void UpdateBoundingBox ();
    // -----------------------------------
    /// \brief Test if vertices make a valid polygon, ie. is convex
    /// \returns "bool" true if valid
    bool IsValid ();

    // -----------------------------------
    /// \brief Interface to test if a point is within any AiBlock
    /// \param "const Real & x" point x position
    /// \param "const Real & y" point y position
    /// \returns "bool" true if inside the bounding box
    virtual bool IsPointInside (const Real& x, const Real& y);

    std::vector<Vector> vertices; ///< the ordered array of vertices

  protected:
    // -----------------------------------
    /// \brief Test if a point is inside of a the polygon
    /// \param "const Real & x" point x position
    /// \param "const Real & y" point y position
    /// \returns "bool" true if inside the polygon
    bool IsPointInside_Polygon (const Real& x, const Real& y);
    // -----------------------------------
    /// \brief Get the distance of a point from a line defined by a point on it and a line normal.
    /// \param "const Vector & line_point" point on the line
    /// \param "const Vector & line_normal" normal to the line
    /// \param "const Vector & test_point" point to test against the line
    /// \returns "BUG::Real" the distance to the line, positive if on same side as the normal.
    Real GetDistanceFromLine (const Vector& line_point, const Vector& line_normal, const Vector& test_point);

    // -----------------------------------
    /// \brief calculate the normals of each line. This allows for faster distance from line calculations
    void CalculateNormals ();

    std::vector<Vector> normals; ///< normal to the each line between points
    bool positive_order; ///< order in which points are stored, clockwise or anti-clockwise. This alters the direction of the line normals.
  };
}

#endif