/// \file AiPathfinder.cpp
/// \date 2008/02/19
/// \author Peter Dodds
/// \brief 
/// \note Copyright (C) 2008 - All Rights Reserved
//------------------------------------------------------------------------------

#include "AiPathfinder.h"
#include "AiPathNode.h"

#include <iostream>
#include <sstream>
#include <fstream>
#include <assert.h>
#include <DebugUtl.h>

namespace BUG
{
  unsigned AiPathfinder::_nextID = 1; // set _nextID to start condition

  //------------------------------------------------ AiPathfinder::AiPathfinder
  AiPathfinder::AiPathfinder()
  {

  }

  //------------------------------------------------ AiPathfinder::~AiPathfinder
  AiPathfinder::~AiPathfinder()
  {
    std::vector<AiPathNode*>::iterator it = m_pathnodes.begin();
    std::vector<AiPathNode*>::iterator endit = m_pathnodes.end();
    for(;it!=endit;++it)
      delete (*it);
  }

  //------------------------------------------------ AiPathfinder::NewNode
  AiPathNode* AiPathfinder::NewNode()
  {
    AiPathNode* nNode = new AiPathNode;
    nNode->id = GetUniqueID();
    m_pathnodes.push_back(nNode);
    return nNode;
  }

  //------------------------------------------------ AiPathfinder::DeleteNode
  bool AiPathfinder::DeleteNode( AiPathNode* _n )
  {
    bool found = false;
    std::vector<AiPathNode*>::iterator it = m_pathnodes.begin();
    for(;it!=m_pathnodes.end();++it)
    {
      if (*it == _n)
      {
        m_pathnodes.erase(it);
        found = true;
        break;
      }
    }

    if (found)
    {
      std::vector<BUG::AiPathNode*>::iterator it = _n->m_successors.begin();
      for (;it!=_n->m_successors.end();++it)
      {
        // assume the node network is set up correctly, and has no duplicate successors
        std::vector<BUG::AiPathNode*>::iterator s_it = (*it)->m_successors.begin();
        for (;s_it!=(*it)->m_successors.end();++s_it)
        {
          if (*s_it == _n)
          {
            // remove _n from the successors
            (*it)->m_successors.erase(s_it);
            break;
          }
        }
      }
    
      delete _n;
    }

    return found;
  }

  //------------------------------------------------ AiPathfinder::ClearNodes
  void AiPathfinder::ClearNodes()
  {
    std::vector<AiPathNode*>::iterator it = m_pathnodes.begin();
    std::vector<AiPathNode*>::iterator endit = m_pathnodes.end();
    for(;it!=endit;++it)
      delete (*it);

    m_pathnodes.clear();
    _nextID = 1;
  }

  //------------------------------------------------ AiPathfinder::LinkNodes
  bool AiPathfinder::LinkNodes( AiPathNode* _a, AiPathNode* _b )
  {
    // check if _b is already a successor of _a
    bool valid = true, success = false;

    // check _b against _a
    std::vector<BUG::AiPathNode*>::iterator it = _a->m_successors.begin();
    for (;it!=_a->m_successors.end();++it)
    {
      if (*it == _b)
      {
        valid = false;
        break;
      }
    }

    if (valid)
    {
      // add _b to _a
      _a->m_successors.push_back(_b);
      success = true;
    }
    else
      valid = true;

    // check _a against _b
    it = _b->m_successors.begin();
    for (;it!=_b->m_successors.end();++it)
    {
      if (*it == _a)
      {
        valid = false;
        break;
      }
    }

    if (valid)
    {
      // add _a to _b
      _b->m_successors.push_back(_a);
      success = true;
    }

    return success;
  }

  //------------------------------------------------ AiPathfinder::GenerateNodeNetwork
  void AiPathfinder::GenerateNodeNetwork( const unsigned& maxLinks /*= 5*/, const Real& maxSpace /*= 40*/ )
  {
    Real maxSpace_sqrd = maxSpace*maxSpace;

    std::vector<AiPathNode*>::iterator it_pN = m_pathnodes.begin();
    std::vector<AiPathNode*>::iterator endit_pN = m_pathnodes.end();

    for (;it_pN!=endit_pN;++it_pN)
    {
      std::vector<AiPathNode*>::iterator it;
      std::vector<AiPathNode*>::iterator endit;
      for (unsigned i=0; (*it_pN)->m_successors.size() < maxLinks; ++i)
      {
        // set first state
        AiPathNode* bNode;
        Real best, _x, _y;


        // set best distance to the first valid element
        bool _ok = false;
        it = m_pathnodes.begin(); endit = m_pathnodes.end();
        while (!_ok)
        {
          // check iterator is still in scope
          if (it == endit)
            return;

          _ok = true;

          // _N is invalid successor
          if (*it == (*it_pN))
          {
            _ok = false;
            ++it;
            continue;
          }

          std::vector<AiPathNode*>::iterator s_it = (*it_pN)->m_successors.begin();
          std::vector<AiPathNode*>::iterator s_end = (*it_pN)->m_successors.end();
          for (;s_it!=s_end;++s_it)
          {
            // if node is already present on successor list; it is not valid
            if (*s_it == *it)
            {
              _ok = false;
              ++it;
              break;
            }
          }
        }

        _x = (*it)->m_pos.x - (*it_pN)->m_pos.x; _y = (*it)->m_pos.y - (*it_pN)->m_pos.y;
        best = _x*_x + _y*_y;
        bNode = 0;
        //++it;

        // test if the first element is the best

        for (;it!=endit;++it)
        {
          if(*it == (*it_pN) || (*it)->m_successors.size() >= maxLinks)
            continue;

          bool invalid = false;
          std::vector<AiPathNode*>::iterator s_it = (*it_pN)->m_successors.begin();
          std::vector<AiPathNode*>::iterator s_end = (*it_pN)->m_successors.end();
          for (;s_it!=s_end;++s_it)
          {
            // if node is already present on successor list; it is not valid
            if (*s_it == *it)
            {
              invalid = true;
              break;
            }
          }

          if (invalid)
            continue;

          _x = (*it)->m_pos.x - (*it_pN)->m_pos.x; _y = (*it)->m_pos.y - (*it_pN)->m_pos.y;
          Real test = _x*_x + _y*_y;
          if ( test < maxSpace_sqrd && (best > test || FCOMPARE(best,test)) )
          {
            best = test;
            bNode = *it;
          }
        }

        if (bNode == 0)
          break;

        (*it_pN)->m_successors.push_back(bNode);
        bNode->m_successors.push_back(*it_pN);
      }
    }
  }

  //------------------------------------------------ AiPathfinder::LoadNodeNetwork
  bool AiPathfinder::LoadNodeNetwork( const char* filename )
  {
    ClearNodes();

    std::ifstream fs;
    fs.open(filename);
    if (!fs.is_open())
      return false;

    // get all the nodes
    while (!fs.eof())
    {
      char buffer[256];

      fs.getline (buffer,256);
      if (buffer[0] == 'd')
      {
        unsigned id, toll;
        int x, y;
        if (sscanf (buffer, "d %d\t %d %d %d", &id, &x, &y, &toll) != 4)
          continue;

        AiPathNode* nnode = NewNode();
        nnode->id = id;
        nnode->m_pos.x = x;
        nnode->m_pos.y = y;
        nnode->m_toll = toll;
      }
    }

    // set next node to largest id
    unsigned last_id = 1;

    std::vector<AiPathNode*>::iterator it = m_pathnodes.begin();
    std::vector<AiPathNode*>::iterator endit = m_pathnodes.end();
    for (;it!=endit;++it)
    {
      if ((*it)->id > last_id)
        last_id = (*it)->id;
    }

    _nextID = ++last_id;




    // rewind and read how to link the nodes
    fs.clear();
    fs.seekg (0);

    it = m_pathnodes.begin();

    while (!fs.eof())
    {
      char buffer[256];

      fs.getline (buffer,256);
      if (buffer[0] == 's')
      {
        std::string buf;  // Have a buffer string
        std::stringstream ss(std::string(buffer), std::stringstream::in | std::stringstream::out);  // Insert the string into a stream

        ss >> buf;
        //assert(buf[0] == 's');

        unsigned node_id;
        ss >> node_id;

        // find node id
        // if (it) is not correct, search for the correct node
        if ((*it)->id != node_id)
        {
          it = m_pathnodes.begin();
          for(;it!=endit;++it)
          {
            if ((*it)->id == node_id)
              break;
          }

          // if id wasn't found
          if (it == endit)
            continue;
        }

        unsigned id;
        while (ss >> id)
        {
          std::vector<AiPathNode*>::iterator s_it = m_pathnodes.begin();
          for(;s_it!=endit;++s_it)
          {
            if ((*s_it)->id == id)
            {
              (*it)->m_successors.push_back(*s_it);
              break;
            }
          }
        }

        if (it + 1 != endit)
          ++it;
      }
    }

    fs.close();

    if (m_pathnodes.empty())
      return false;

    return true;
  }

  //------------------------------------------------ AiPathfinder::SaveNodeNetwork
  bool AiPathfinder::SaveNodeNetwork( const char* filename, bool force /*= false*/ )
  {
    // if we don't want to override the file
    if (!force)
    {
      // try and open the file to see if it already exists
      std::ifstream _fs;
      _fs.open(filename);
      if (_fs.is_open())
      {
        _fs.close();
        DBUG__(DOUT << "File already exists, will not override" << "\n");
        return false;
      }
    }

    // open the file for writing
    std::ofstream fs;
    fs.open(filename);
    if (!fs.is_open())
      return false;

    // write header?

    // write out each Node's data
    std::vector<AiPathNode*>::iterator it = m_pathnodes.begin();
    std::vector<AiPathNode*>::iterator endit = m_pathnodes.end();
    for(;it!=endit;++it)
    {
      fs << "d " << (*it)->id << "\t " 
        << (*it)->m_pos.x << " " << (*it)->m_pos.y << " "
        << (*it)->m_toll << " ";
      WriteNodeExtraData(fs,*it);
      fs << std::endl;
    }

    // write out links.
    it = m_pathnodes.begin();
    for(;it!=endit;++it)
    {
      fs << "s " << (*it)->id << "\t";

      std::vector<AiPathNode*>::iterator s_it = (*it)->m_successors.begin();
      std::vector<AiPathNode*>::iterator s_end = (*it)->m_successors.end();
      for (;s_it!=s_end;++s_it)
        fs << " " << (*s_it)->id;

      fs << std::endl;
    }

    fs.close();
    
    return true;
  }

  //------------------------------------------------ AiPathfinder::WriteNodeExtraData
  void AiPathfinder::WriteNodeExtraData( std::ostream& os, AiPathNode* node )
  {
    
  }

  //------------------------------------------------ AiPathfinder::GetClosestNode
  AiPathNode* AiPathfinder::GetClosestNode( const Real& _x, const Real& _y )
  {
    if (m_pathnodes.empty())
      return 0;

    AiPathNode* _returnNode = 0;
    Real distance;

    std::vector<AiPathNode*>::iterator it = m_pathnodes.begin();
    std::vector<AiPathNode*>::iterator endit = m_pathnodes.end();

    // distance = sqrt (x*x + y*y); x = n.x - _x; y = n.y - _y
    // but in comparison there is no need to use sqrt ()
    // assume that the first _N is the closest, then check if correct
    int _tX = (*it)->m_pos.x - _x, _tY = (*it)->m_pos.y - _y;
    distance = _tX*_tX + _tY*_tY;
    _returnNode = (*it);
    ++it;

    Real testDist;
    for (;it!=endit;++it)
    {
      _tX = (*it)->m_pos.x - _x; _tY = (*it)->m_pos.y - _y;
      testDist = _tX*_tX + _tY*_tY;
      if (testDist < distance)
      {
        distance = testDist;
        _returnNode = (*it);
      }
    }

    return _returnNode;
  }

  //------------------------------------------------ AiPathfinder::AStar_GetSuccessors
  void AiPathfinder::AStar_GetSuccessors( AiPathNode* _N )
  {
    // this requires an understanding of how the AiPathNodes are laid out.
    // each _N could store pointers to connected nodes, but by default I must assum there is no order

    _successors.clear();
    _successors = _N->m_successors;
  }

  //------------------------------------------------ AiPathfinder::AStar_CalculateCost
  unsigned AiPathfinder::AStar_CalculateCost( AiPathNode* current, AiPathNode* next )
  {
    return current->aStar_cost + next->m_toll;
  }

  //------------------------------------------------ AiPathfinder::AStar_EstimateGoalDistance
  unsigned AiPathfinder::AStar_EstimateGoalDistance( AiPathNode* node, AiPathNode* goal )
  {
    // heuristic function to estimate the number of steps to the goal.
    // The behavior of the A* pathfinding is altered in the use of function.
    // this requires an understanding of how the AiPathNodes are laid out.
    
    // this is a simple (and bad) implementation. just calculating the straight line distance and divide by 2.

    Real x = goal->m_pos.x - node->m_pos.x;
    Real y = goal->m_pos.y - node->m_pos.y;

    return (unsigned)(BUG_SQRT(x*x + y*y));
  }


  //------------------------------------------------ AiPathfinder::CalculatePath
  BUG::AiPath AiPathfinder::CalculatePath( AiPathNode* start, AiPathNode* end )
  {
    if (start == end)
      return AiPath ();

    std::vector<AiPathNode*>::iterator it;
    std::vector<AiPathNode*>::iterator endit;
    bool solved = false;

    _open.clear();
    _closed.clear();

    start->aStar_parent = 0;
    start->aStar_cost = 0;
    start->aStar_goalDistance = AStar_EstimateGoalDistance(start,end);
    _open.push_back(start);
    while (!_open.empty())
    {
      // get the lowest f() _N on open list.
      // f = cost (g) + goalDistance (h)
      // assume that the first is lowest, then test assumtion.
      it = _open.begin(); endit = _open.end();
      AiPathNode* cur = (*it);
      ++it;
      for (;it!=endit;++it)
      {
        if (((*it)->aStar_cost + (*it)->aStar_goalDistance) < (cur->aStar_cost + cur->aStar_goalDistance))
          cur = (*it);
      }

      if (cur == end)
      {
        solved = true;
        break;
      }

      AStar_GetSuccessors(cur);
      it = _successors.begin(); endit = _successors.end();
      for (;it!=endit;++it)
      {
        unsigned succ_cost = AStar_CalculateCost(cur,*it);
        std::vector<AiPathNode*>::iterator open_it = _open.begin();
        for (;open_it!=_open.end();++open_it)
        {
          if (*open_it == *it)
            break;
        }
        // if successor was found on the open list, but succ_cost was greater
        //    Discard the successor and continue
        if (open_it != _open.end() && (*open_it)->aStar_cost < succ_cost)
          continue;


        std::vector<AiPathNode*>::iterator closed_it = _closed.begin();
        for (;closed_it!=_closed.end();++closed_it)
        {
          if (*closed_it == *it)
            break;
        }
        // if successor was found on the closed list, but succ_cost was greater
        //    Discard the successor and continue
        if (closed_it != _closed.end() && (*closed_it)->aStar_cost < succ_cost)
          continue;


        // remove instances of successor from open and closed lists
        if (open_it != _open.end())
          _open.erase(open_it);

        if (closed_it != _closed.end())
          _closed.erase(closed_it);

        (*it)->aStar_cost = succ_cost;
        (*it)->aStar_parent = cur;
        (*it)->aStar_goalDistance = AStar_EstimateGoalDistance(*it,end);
        _open.push_back(*it);
      }

      // remove current from open and add to to closed
      it = _open.begin(); endit = _open.end();
      for (;it!=endit;++it)
      {
        if (cur == (*it))
        {
          _open.erase(it);
          break;
        }
      }
      _closed.push_back(cur);
    }

    AiPath _r;

    if (solved)
    {
      AiPathNode* trace = end;
      _r.nodes.insert(_r.nodes.begin(),trace);

      while(true)
      {
        trace = trace->aStar_parent;
        if (trace == 0)
        {
          // somthing's gone wrong
          _r.nodes.clear();
          break;
        }

        _r.nodes.insert(_r.nodes.begin(),trace);
        if (trace == start)
          break;
      }
    }

    return _r;
  }
}