/// \file AiPathfinder.h
/// \date 2008/02/19
/// \author Peter Dodds
/// \brief 
/// \note Copyright (C) 2008 - All Rights Reserved
//------------------------------------------------------------------------------

#ifdef _MSC_VER
#pragma once
#endif

#ifndef _BUG_AiPathfinder_h_
#define _BUG_AiPathfinder_h_

#include "AiPath.h"
#include "AiDef.h"
#include <vector>
#include <ostream>

namespace BUG
{
  class AiPathNode;

	/// \brief 
	class AiPathfinder
	{
	public:
		//---------------------------------------
		/// \brief Default Constructor
		AiPathfinder();
		//---------------------------------------
		/// \brief Destructor
		virtual ~AiPathfinder();
		
    // -----------------------------------
    /// \brief create a new pathnode
    /// \returns "AiPathNode*" new node
    AiPathNode* NewNode ();
    // -----------------------------------
    /// \brief delete a specific pathnode
    /// \param "AiPathNode * _n" node to delete
    /// \returns "bool" true is the node was deleted
    bool DeleteNode (AiPathNode* _n);
    // -----------------------------------
    /// \brief delete all nodes
    void ClearNodes ();

    // -----------------------------------
    /// \brief link two nodes so they can be traversed correctly
    /// \param "AiPathNode * _a" first node
    /// \param "AiPathNode * _b" second node
    /// \returns "bool" true if they are connected
    bool LinkNodes (AiPathNode* _a, AiPathNode* _b);

    // -----------------------------------
    /// \brief Generate an entire network of nodes
    /// \param "const unsigned & maxLinks" maximum number of link to a single node
    /// \param "const Real & maxSpace" maximum spacing between two linked nodes
    virtual void GenerateNodeNetwork (const unsigned& maxLinks = 5, const Real& maxSpace = 40);
    // -----------------------------------
    /// \brief load a node network file
    /// \param "const char * filename" file to load from
    /// \returns "bool" true if loaded successfully
    virtual bool LoadNodeNetwork (const char* filename);
    // -----------------------------------
    /// \brief save a node network file
    /// \param "const char * filename" file to save to
    /// \param "bool force" force override of file if it exists
    /// \returns "bool" true if file was saved
    bool SaveNodeNetwork (const char* filename, bool force = false);

  protected:
    virtual void WriteNodeExtraData (std::ostream& os, AiPathNode* node);

  public:
    // -----------------------------------
    /// \brief Find the closest node
    /// \param "const Real & _x" x test position
    /// \param "const Real & _y" y test position
    /// \returns "AiPathNode*" closest node
    virtual AiPathNode* GetClosestNode (const Real& _x, const Real& _y);

    // -----------------------------------
    /// \brief Calculate the path between two path nodes 
    /// \param "AiPathNode * start"
    /// \param "AiPathNode * end"
    /// \returns "BUG::AiPath" the path
    virtual AiPath CalculatePath (AiPathNode* start, AiPathNode* end);

  protected:
    // -----------------------------------
    /// \brief get the successors to a pathnode, this uses GetClosestNode
    /// \param "AiPathNode * _N" 
    virtual void AStar_GetSuccessors (AiPathNode* _N);
    // -----------------------------------
    /// \brief calculate cost of a moving to a pathnode
    /// \param "AiPathNode * current" the current location
    /// \param "AiPathNode * next" the next location
    /// \returns "unsigned" the cost
    virtual unsigned AStar_CalculateCost (AiPathNode* current, AiPathNode* next);
    // -----------------------------------
    /// \brief heuristic function to estimate the distance to a goal
    /// \param "AiPathNode * node" current node
    /// \param "AiPathNode * goal" goal node
    /// \returns "unsigned" the estimated distance
    virtual unsigned AStar_EstimateGoalDistance (AiPathNode* node, AiPathNode* goal);

    std::vector<AiPathNode*> _open, _closed, _successors; ///< A* specific variables

  public:
    std::vector<AiPathNode*> m_pathnodes; ///< list of all pathnodes in the game.

  private:
    // -----------------------------------
    /// \brief get a unique ID for pathnodes.
    /// \returns "unsigned"
    unsigned GetUniqueID ()
    { 
      return _nextID++; 
    }
    static unsigned _nextID;
	};
}

#endif